import logging
from abc import abstractmethod

from webpub_manifest_parser.core.analyzer import BaseAnalyzer, BaseAnalyzerError
from webpub_manifest_parser.core.parsers import (
    ArrayParser,
    TypeParser,
    ValueParserError,
    find_parser,
)
from webpub_manifest_parser.core.properties import BaseArrayProperty, PropertiesGrouping
from webpub_manifest_parser.utils import encode, first_or_default, is_string


class SyntaxAnalyzerError(BaseAnalyzerError):
    """Exception raised in the case of syntax errors."""


class MissingPropertyError(SyntaxAnalyzerError):
    """Exception raised in the case of a missing required property."""

    def __init__(self, node, node_property, message=None, inner_exception=None):
        """Initialize a new instance of MissingPropertyError class.

        :param node: AST node where the error was found
        :type node: webpub_manifest_parser.core.ast.Node

        :param node_property: AST node's property associated with the error
        :type node_property: webpub_manifest_parser.core.properties.Property

        :param message: (Optional) Message
        :type message: Optional[str]

        :param inner_exception: (Optional) inner exception
        :type inner_exception: Optional[Exception]
        """
        if message is None:
            message = "{}'s required property {} is missing".format(
                node.__class__, node_property.key
            )

        super().__init__(
            node,
            node_property,
            message,
            inner_exception,
        )


class SyntaxAnalyzer(BaseAnalyzer):
    """Base class for syntax analyzers checking the base grammar rules of RWPM and parsing raw JSON into AST."""

    CONTEXT = "@context"
    METADATA = "metadata"
    METADATA_IDENTIFIER = "identifier"
    LINKS = "links"

    def __init__(self):
        """Initialize a new instance of SyntaxParser class."""
        super().__init__()

        self._logger = logging.getLogger(__name__)

    @abstractmethod
    def _create_manifest(self):
        """Create a new manifest. The method should be overridden in child classes.

        :return: Manifest-like object
        :rtype: Manifestlike
        """
        raise NotImplementedError()

    def _get_property_value(self, json_content, object_property):
        """Extract the property's value from JSON object.

        :param json_content: JSON object
        :type json_content: Dict

        :param object_property: Object's property
        :type object_property: properties.Property

        :return: Property's value
        :rtype: Any
        """
        self._logger.debug(f"Started extracting {object_property} property")

        if isinstance(json_content, dict):
            property_value = json_content.get(object_property.key, None)
        else:
            property_value = json_content

        self._logger.debug(
            "Finished extracting {} property: {}".format(
                object_property, encode(property_value)
            )
        )

        return property_value

    def _parse_nested_object(self, property_value, object_property):
        """Parse nested object(s) (if any) and return the result of parsing.

        :param property_value: Raw property's value (probably) containing nested object(s)
        :type property_value: Any

        :param object_property: Object's property
        :type object_property: properties.Property

        :return: Nested object's value
        :rtype: Any
        """
        if property_value is None:
            return property_value

        self._logger.debug(f"Started looking for nested property {object_property}")

        type_parsers_result = find_parser(object_property.parser, TypeParser)

        self._logger.debug(f"Found the following type parsers: {type_parsers_result}")

        found = False

        for parent_parser, type_parser in type_parsers_result:
            if isinstance(parent_parser, ArrayParser) and isinstance(
                property_value, list
            ):
                processed_items = []

                for item in property_value:
                    processed_item = self._parse_object(item, type_parser.type)
                    processed_items.append(processed_item)

                found = True
                property_value = processed_items
                break
        else:
            for parent_parser, type_parser in type_parsers_result:
                if not isinstance(parent_parser, ArrayParser):
                    found = True
                    property_value = self._parse_object(
                        property_value, type_parser.type
                    )
                    break

        if found:
            self._logger.debug(
                "Finished parsing nested property {}: {}".format(
                    object_property, encode(property_value)
                )
            )
        else:
            self._logger.debug(f"Property {object_property} is not nested")

        return property_value

    @staticmethod
    def _format_property_value(property_value, object_property):
        """Format the property's value according to its format.

        For example, values of list properties have to be a collection of the type specified in the property.
        """
        if isinstance(object_property, BaseArrayProperty) and not isinstance(
            property_value, object_property.list_type
        ):
            if property_value is not None:
                property_value = object_property.list_type(
                    property_value
                    if isinstance(property_value, list)
                    else [property_value]
                )
            else:
                property_value = object_property.list_type()

        return property_value

    def _set_scalar_value(self, json_content, ast_object):
        """Parse a scalar string value and initialize an object's property with it.

        :param json_content: Scalar string value containing a required object's property
        :type json_content: str

        :param ast_object: AST object
        :type ast_object: Node
        """
        required_object_properties = PropertiesGrouping.get_required_class_properties(
            ast_object.__class__
        )

        if len(required_object_properties) != 1:
            raise SyntaxAnalyzerError(
                ast_object,
                None,
                "There are {} required properties in {} but only a single value ({}) was provided".format(
                    len(required_object_properties),
                    encode(ast_object),
                    json_content,
                ),
            )

        required_object_property_name, required_object_property = first_or_default(
            required_object_properties
        )

        self._set_property_value(
            ast_object,
            required_object_property_name,
            required_object_property,
            json_content,
        )

        # We need to initialize other properties with default values
        self._set_non_scalar_value(None, ast_object, {required_object_property_name})

    def _set_non_scalar_value(
        self, json_content, ast_object, excluded_property_names=None
    ):
        """Parse a dictionary and initialize object's properties with its values.

        :param json_content: Dictionary containing property values
        :type json_content: Dict

        :param ast_object: AST object
        :type ast_object: Node

        :param excluded_property_names: Set of property names to exclude from consideration
        :type excluded_property_names: Set
        """
        ast_object_properties = PropertiesGrouping.get_class_properties(
            ast_object.__class__
        )

        for object_property_name, object_property in ast_object_properties:
            if (
                excluded_property_names
                and object_property_name in excluded_property_names
            ):
                continue

            with self._record_errors():
                property_value = self._get_property_value(json_content, object_property)
                property_value = self._parse_nested_object(
                    property_value, object_property
                )

                self._set_property_value(
                    ast_object, object_property_name, object_property, property_value
                )

    def _set_property_value(
        self, ast_object, object_property_name, object_property, property_value
    ):
        """Set the value of the specified property.

        :param ast_object: AST object
        :type ast_object: Node

        :param object_property_name: Name of the property
        :type object_property_name: str

        :param object_property: Object's property
        :type object_property: Property

        :param property_value: Value to be set
        :type property_value: Any
        """
        self._logger.debug(
            "Property '{}' has the following value: {}".format(
                object_property.key, encode(property_value)
            )
        )

        if property_value is None and object_property.default_value is not None:
            property_value = object_property.default_value

        if object_property.required and property_value is None:
            raise MissingPropertyError(ast_object, object_property)

        if property_value is not None:
            try:
                property_value = object_property.parser.parse(property_value)
            except ValueParserError as error:
                self._logger.error(
                    f"Error while parsing {property_value} for {object_property.key}, falling back to default"
                )
                # First, fallback to the default value for the property, then re-raise
                property_value = self._format_property_value(
                    object_property.default_value, object_property
                )
                setattr(ast_object, object_property_name, property_value)
                raise SyntaxAnalyzerError(
                    ast_object, object_property, error.error_message, error
                )

        property_value = self._format_property_value(property_value, object_property)

        setattr(ast_object, object_property_name, property_value)

    def _parse_object(self, json_content, cls):
        """Parse RWPM's object JSON into a corresponding AST object.

        :param json_content: Dictionary containing object's JSON
        :type json_content: Dict

        :param cls: Object's class
        :type cls: Type

        :return: Node object
        :rtype: Node
        """
        self._logger.debug(f"Started parsing {cls} object")

        extended_cls = cls.get_extension()
        ast_object = extended_cls()

        with self._record_errors():
            if is_string(json_content):
                self._set_scalar_value(json_content, ast_object)
            elif isinstance(json_content, (list, dict)):
                self._set_non_scalar_value(json_content, ast_object)

        self._logger.debug(f"Finished parsing {cls} object: {ast_object}")

        return ast_object

    def analyze(self, manifest_json):
        """Parse JSON file into RWPM AST.

        :param manifest_json: RWPM-compatible manifest
        :type manifest_json: Dict

        :return: RWPM AST
        :rtype: ManifestLike
        """
        self._logger.debug(f"Started analyzing {manifest_json}")

        self.context.reset()

        manifest = self._create_manifest()
        manifest = self._parse_object(manifest_json, manifest.__class__)

        self._logger.debug(f"Finished analyzing {manifest_json}: {manifest}")

        return manifest



import logging
from functools import partial

from multipledispatch import dispatch

from webpub_manifest_parser.core.ast import (
    Collection,
    CollectionList,
    CompactCollection,
    Link,
    LinkList,
    Manifestlike,
    Metadata,
)
from webpub_manifest_parser.core.semantic import LinkSemanticError, SemanticAnalyzer

READING_ORDER_SUBCOLLECTION_LINK_MISSING_TYPE_PROPERTY_ERROR = partial(
    LinkSemanticError,
    message="readingOrder subcollection's link '{0}' does not have a required 'type' property",
)
RESOURCES_SUBCOLLECTION_MISSING_LINK_TYPE_PROPERTY_ERROR = partial(
    LinkSemanticError,
    message="resources subcollection's link '{0}' does not have a required 'type' property",
)


class RWPMSemanticAnalyzer(SemanticAnalyzer):
    """RWPM semantic analyzer."""

    def __init__(
        self, media_types_registry, link_relations_registry, collection_roles_registry
    ):
        """Initialize a new instance of RWPMSemanticAnalyzer class.

        :param media_types_registry: Media types registry
        :type media_types_registry: python_rwpm_parser.registry.Registry

        :param link_relations_registry: Link relations registry
        :type link_relations_registry: python_rwpm_parser.registry.Registry

        :param collection_roles_registry: Collections roles registry
        :type collection_roles_registry: python_rwpm_parser.registry.Registry
        """
        super().__init__(
            media_types_registry, link_relations_registry, collection_roles_registry
        )

        self._logger = logging.getLogger(__name__)

    @dispatch(Manifestlike)
    def visit(self, node):
        """Perform semantic analysis of the manifest node.

        :param node: Manifest's metadata
        :type node: RWPMManifest
        """
        super().visit(node)

        for link in node.reading_order.links:
            with self._record_errors():
                if link.type is None:
                    raise READING_ORDER_SUBCOLLECTION_LINK_MISSING_TYPE_PROPERTY_ERROR(
                        node=link, node_property=Link.type
                    )

        if node.resources:
            for link in node.resources.links:
                with self._record_errors():
                    if link.type is None:
                        raise RESOURCES_SUBCOLLECTION_MISSING_LINK_TYPE_PROPERTY_ERROR(
                            node=link, node_property=Link.type
                        )

    @dispatch(Metadata)
    def visit(self, node):
        """Perform semantic analysis of the manifest's metadata.

        :param node: Manifest's metadata
        :type node: Metadata
        """
        super().visit(node)

    @dispatch(LinkList)
    def visit(self, node):
        """Perform semantic analysis of the list of links.

        :param node: Manifest's metadata
        :type node: LinkList
        """
        super().visit(node)

    @dispatch(Link)
    def visit(self, node):
        """Perform semantic analysis of the link node.

        :param node: Link node
        :type node: Link
        """
        super().visit(node)

    @dispatch(CollectionList)
    def visit(self, node):
        """Perform semantic analysis of the list of sub-collections.

        :param node: CollectionList node
        :type node: CollectionList
        """
        super().visit(node)

    @dispatch(CompactCollection)
    def visit(self, node):
        """Perform semantic analysis of the compact collection node.

        :param node: Collection node
        :type node: CompactCollection
        """
        super().visit(node)

    @dispatch(Collection)
    def visit(self, node):
        """Perform semantic analysis of the collection node.

        :param node: Collection node
        :type node: Collection
        """
        super().visit(node)



from enum import Enum

from webpub_manifest_parser.core.ast import (
    ArrayOfCollectionsProperty,
    Collection,
    CollectionList,
    CompactCollection,
    CompactCollectionProperty,
    LinkList,
    LinkProperties,
    Manifestlike,
    Node,
    PresentationMetadata,
)
from webpub_manifest_parser.core.parsers import (
    AnyOfParser,
    ArrayParser,
    ObjectParser,
    StringParser,
    TypeParser,
)
from webpub_manifest_parser.core.properties import (
    ArrayProperty,
    DateOrTimeProperty,
    DateTimeProperty,
    EnumProperty,
    IntegerProperty,
    NumberProperty,
    ParsableProperty,
    StringProperty,
    TypeProperty,
    URIProperty,
)
from webpub_manifest_parser.opds2.registry import OPDS2CollectionRolesRegistry


class OPDS2Price(Node):
    """OPDS 2.0 price information."""

    value = NumberProperty("value", required=True, minimum=0)
    currency = EnumProperty(
        "currency",
        required=True,
        items=[
            "AED",
            "AFN",
            "ALL",
            "AMD",
            "ANG",
            "AOA",
            "ARS",
            "AUD",
            "AWG",
            "AZN",
            "BAM",
            "BBD",
            "BDT",
            "BGN",
            "BHD",
            "BIF",
            "BMD",
            "BND",
            "BOB",
            "BOV",
            "BRL",
            "BSD",
            "BTN",
            "BWP",
            "BYN",
            "BZD",
            "CAD",
            "CDF",
            "CHE",
            "CHF",
            "CHW",
            "CLF",
            "CLP",
            "CNY",
            "COP",
            "COU",
            "CRC",
            "CUC",
            "CUP",
            "CVE",
            "CZK",
            "DJF",
            "DKK",
            "DOP",
            "DZD",
            "EGP",
            "ERN",
            "ETB",
            "EUR",
            "FJD",
            "FKP",
            "GBP",
            "GEL",
            "GHS",
            "GIP",
            "GMD",
            "GNF",
            "GTQ",
            "GYD",
            "HKD",
            "HNL",
            "HRK",
            "HTG",
            "HUF",
            "IDR",
            "ILS",
            "INR",
            "IQD",
            "IRR",
            "ISK",
            "JMD",
            "JOD",
            "JPY",
            "KES",
            "KGS",
            "KHR",
            "KMF",
            "KPW",
            "KRW",
            "KWD",
            "KYD",
            "KZT",
            "LAK",
            "LBP",
            "LKR",
            "LRD",
            "LSL",
            "LYD",
            "MAD",
            "MDL",
            "MGA",
            "MKD",
            "MMK",
            "MNT",
            "MOP",
            "MRU",
            "MUR",
            "MVR",
            "MWK",
            "MXN",
            "MXV",
            "MYR",
            "MZN",
            "NAD",
            "NGN",
            "NIO",
            "NOK",
            "NPR",
            "NZD",
            "OMR",
            "PAB",
            "PEN",
            "PGK",
            "PHP",
            "PKR",
            "PLN",
            "PYG",
            "QAR",
            "RON",
            "RSD",
            "RUB",
            "RWF",
            "SAR",
            "SBD",
            "SCR",
            "SDG",
            "SEK",
            "SGD",
            "SHP",
            "SLL",
            "SOS",
            "SRD",
            "SSP",
            "STN",
            "SVC",
            "SYP",
            "SZL",
            "THB",
            "TJS",
            "TMT",
            "TND",
            "TOP",
            "TRY",
            "TTD",
            "TWD",
            "TZS",
            "UAH",
            "UGX",
            "USD",
            "USN",
            "UYI",
            "UYU",
            "UZS",
            "VEF",
            "VES",
            "VND",
            "VUV",
            "WST",
            "XAF",
            "XAG",
            "XAU",
            "XBA",
            "XBB",
            "XBC",
            "XBD",
            "XCD",
            "XDR",
            "XOF",
            "XPD",
            "XPF",
            "XPT",
            "XSU",
            "XTS",
            "XUA",
            "XXX",
            "YER",
            "ZAR",
            "ZMW",
            "ZWL",
        ],
    )


class OPDS2AcquisitionObject(Node):
    """OPDS 2.0 acquisition information."""

    type = StringProperty("type", required=True)
    child = ArrayProperty(
        "child",
        required=False,
        item_parser=TypeParser(
            "webpub_manifest_parser.opds2.ast.OPDS2AcquisitionObject"
        ),
    )


class OPDS2HoldsInformation(Node):
    """OPDS 2.0 holds information."""

    total = IntegerProperty("total", required=False, minimum=0)
    position = IntegerProperty("position", required=False, minimum=0)


class OPDS2CopiesInformation(Node):
    """OPDS 2.0 information about available copies."""

    total = IntegerProperty("total", required=False, minimum=0)
    available = IntegerProperty("available", required=False, minimum=0)


class OPDS2AvailabilityType(Enum):
    """Enumeration of different availability types."""

    AVAILABLE = "available"
    UNAVAILABLE = "unavailable"
    RESERVED = "reserved"
    READY = "ready"


class OPDS2AvailabilityInformation(Node):
    """OPDS 2.0 availability information."""

    state = EnumProperty(
        "state",
        required=True,
        items=[
            OPDS2AvailabilityType.AVAILABLE.value,
            OPDS2AvailabilityType.UNAVAILABLE.value,
            OPDS2AvailabilityType.RESERVED.value,
            OPDS2AvailabilityType.READY.value,
        ],
    )
    since = DateOrTimeProperty("since", required=False)
    until = DateOrTimeProperty("until", required=False)


class OPDS2LinkProperties(LinkProperties):
    """OPDS 2.0 link properties."""

    number_of_items = IntegerProperty("numberOfItems", required=False, minimum=0)
    price = TypeProperty("price", required=False, nested_type=OPDS2Price)
    indirect_acquisition = ArrayProperty(
        "indirectAcquisition",
        required=False,
        item_parser=TypeParser(OPDS2AcquisitionObject),
    )
    holds = TypeProperty("holds", required=False, nested_type=OPDS2HoldsInformation)
    copies = TypeProperty("copies", required=False, nested_type=OPDS2CopiesInformation)
    availability = TypeProperty(
        "availability", required=False, nested_type=OPDS2AvailabilityInformation
    )


LinkProperties.extensions = (OPDS2LinkProperties,)


class TitleProperty(ParsableProperty):
    """Property containing arbitrary values."""

    PARSER = AnyOfParser(
        [StringParser(), ArrayParser(StringParser()), ObjectParser(StringParser())]
    )


class OPDS2FeedMetadata(Node):
    """OPDS 2.x feed metadata."""

    identifier = URIProperty("identifier", required=False)
    type = URIProperty("@type", required=False)
    title = TitleProperty("title", required=True)
    subtitle = TitleProperty("subtitle", required=False)
    modified = DateTimeProperty("modified", required=False)
    description = StringProperty("description", required=False)
    items_per_page = IntegerProperty(
        "itemsPerPage", required=False, exclusive_minimum=0
    )
    current_page = IntegerProperty("currentPage", required=False, exclusive_minimum=0)
    number_of_items = IntegerProperty("numberOfItems", required=False, minimum=0)

    def __init__(
        self,
        title=None,
        _type=None,
        subtitle=None,
        modified=None,
        description=None,
        items_per_page=None,
        current_page=None,
        number_of_items=None,
    ):
        """Initialize a new instance of OPDS2FeedMetadata class."""
        super().__init__()

        self.title = title
        self.type = _type
        self.subtitle = subtitle
        self.modified = modified
        self.description = description
        self.items_per_page = items_per_page
        self.current_page = current_page
        self.number_of_items = number_of_items


class OPDS2Publication(Collection):
    """OPDS 2.0 publication."""

    images = CompactCollectionProperty(
        "images", required=True, role=OPDS2CollectionRolesRegistry.IMAGES
    )

    def __init__(self, metadata=None, links=None, images=None):
        """Initialize a new instance of OPDS2Publication class."""
        super().__init__()

        if metadata and not isinstance(metadata, PresentationMetadata):
            raise ValueError(
                "Argument 'metadata' must be an instance of {}".format(
                    PresentationMetadata
                )
            )
        if links and not isinstance(links, LinkList):
            raise ValueError(f"Argument 'links' must be an instance of {LinkList}")
        if images and not isinstance(images, LinkList):
            raise ValueError(f"Argument 'images' must be an instance of {LinkList}")

        self.metadata = metadata
        self.links = links
        self.images = images

    def __hash__(self):
        """Calculate the hash.

        :return: Hash
        :rtype: int
        """
        return hash((super().__hash__(), self.images))


class OPDS2Facet(Collection):
    """OPDS 2.0 facet."""

    metadata = TypeProperty("metadata", required=False, nested_type=OPDS2FeedMetadata)


class OPDS2Navigation(CompactCollection):
    """OPDS 2 navigation for the catalog using links."""


class OPDS2Group(Collection):
    """OPDS 2.0 group."""

    metadata = TypeProperty("metadata", required=False, nested_type=OPDS2FeedMetadata)
    publications = ArrayOfCollectionsProperty(
        "publications",
        required=False,
        role=OPDS2CollectionRolesRegistry.PUBLICATIONS,
        collection_type=OPDS2Publication,
    )
    navigation = CompactCollectionProperty(
        "navigation",
        required=False,
        role=OPDS2CollectionRolesRegistry.NAVIGATION,
        collection_class=OPDS2Navigation,
    )

    def __init__(self, metadata=None, publications=None, navigation=None):
        """Initialize a new instance of OPDS2Group class."""
        super().__init__()

        if metadata and not isinstance(metadata, OPDS2FeedMetadata):
            raise ValueError(
                "Argument 'metadata' must be an instance of {}".format(
                    OPDS2FeedMetadata
                )
            )
        if publications and not isinstance(publications, CollectionList):
            raise ValueError(
                "Argument 'publications' must be an instance of {}".format(
                    CollectionList
                )
            )
        if navigation and not isinstance(navigation, OPDS2Navigation):
            raise ValueError(
                "Argument 'navigation' must be an instance of {}".format(
                    OPDS2Navigation
                )
            )

        self.metadata = metadata
        self.publications = publications
        self.navigation = navigation


class OPDS2Feed(Manifestlike):
    """OPDS 2.x feed."""

    metadata = TypeProperty("metadata", required=True, nested_type=OPDS2FeedMetadata)
    publications = ArrayOfCollectionsProperty(
        "publications",
        required=False,
        role=OPDS2CollectionRolesRegistry.PUBLICATIONS,
        collection_type=OPDS2Publication,
    )
    navigation = CompactCollectionProperty(
        "navigation",
        required=False,
        role=OPDS2CollectionRolesRegistry.NAVIGATION,
        collection_class=OPDS2Navigation,
    )
    facets = ArrayOfCollectionsProperty(
        "facets",
        required=False,
        role=OPDS2CollectionRolesRegistry.FACETS,
        collection_type=OPDS2Facet,
    )
    groups = ArrayOfCollectionsProperty(
        "groups",
        required=False,
        role=OPDS2CollectionRolesRegistry.GROUPS,
        collection_type=OPDS2Group,
    )

    def __init__(
        self,
        metadata=None,
        links=None,
        publications=None,
        navigation=None,
        facets=None,
        groups=None,
    ):
        """Initialize a new instance of OPDS2Feed class."""
        super().__init__()

        if metadata and not isinstance(metadata, OPDS2FeedMetadata):
            raise ValueError(
                "Argument 'metadata' must be an instance of {}".format(
                    OPDS2FeedMetadata
                )
            )
        if links and not isinstance(links, LinkList):
            raise ValueError(f"Argument 'links' must be an instance of {LinkList}")
        if publications and not isinstance(publications, CollectionList):
            raise ValueError(
                "Argument 'publications' must be an instance of {}".format(
                    CollectionList
                )
            )
        if navigation and not isinstance(navigation, OPDS2Navigation):
            raise ValueError(
                "Argument 'navigation' must be an instance of {}".format(
                    OPDS2Navigation
                )
            )
        if facets and not isinstance(facets, CollectionList):
            raise ValueError(
                f"Argument 'facets' must be an instance of {CollectionList}"
            )
        if groups and not isinstance(groups, CollectionList):
            raise ValueError(
                f"Argument 'groups' must be an instance of {CollectionList}"
            )

        self.metadata = metadata
        self.links = links
        self.publications = publications
        self.navigation = navigation
        self.facets = facets
        self.groups = groups



from webpub_manifest_parser.core.syntax import SyntaxAnalyzer
from webpub_manifest_parser.opds2.ast import OPDS2Feed


class OPDS2SyntaxAnalyzer(SyntaxAnalyzer):
    """OPDS 2.0 syntax analyzer."""

    def _create_manifest(self):
        """Create a new OPDS 2.0 manifest.

        :return: OPDS 2.0 manifest
        :rtype: Manifestlike
        """
        return OPDS2Feed()



from webpub_manifest_parser.odl.ast import ODLFeed
from webpub_manifest_parser.opds2.syntax import OPDS2SyntaxAnalyzer


class ODLSyntaxAnalyzer(OPDS2SyntaxAnalyzer):
    """OPDS 2.0 syntax analyzer."""

    def _create_manifest(self):
        """Create a new OPDS 2.0 manifest.

        :return: OPDS 2.0 manifest
        :rtype: Manifestlike
        """
        return ODLFeed()



from webpub_manifest_parser.core.ast import CompactCollectionProperty, Manifestlike
from webpub_manifest_parser.core.properties import ArrayOfStringsProperty
from webpub_manifest_parser.rwpm.registry import RWPMCollectionRolesRegistry


class RWPMManifest(Manifestlike):
    """Readium Web Publication Manifest."""

    # https://github.com/readium/webpub-manifest#22-metadata
    DEFAULT_CONTEXT = "https://readium.org/webpub-manifest/context.jsonld"

    def __init__(
        self,
        metadata=None,
        links=None,
        reading_order=None,
        context=None,
        resources=None,
        toc=None,
    ):
        """Initialize a new instance of RWPMManifest class.

        :param metadata: RWPM metadata
        :type metadata: webpub_manifest_parser.core.ast.Metadata

        :param links: Links
        :type links: webpub_manifest_parser.core.ast.LinkList

        :param reading_order: Reading order sub-collection
        :type reading_order: webpub_manifest_parser.core.ast.CompactCollection

        :param resources: (Optional) Resources sub-collection
        :type resources: webpub_manifest_parser.core.ast.CompactCollection

        :param context: (Optional) RWPM manifest's context
        :type context: list

        :param toc: (Optional) TOC sub-collection
        :type toc: webpub_manifest_parser.core.ast.CompactCollection
        """
        super().__init__()

        self.metadata = metadata
        self.links = links
        self.reading_order = reading_order
        self.context = context
        self.resources = resources
        self.toc = toc

    context = ArrayOfStringsProperty(
        "@context", required=False, unique_items=True, default_value=DEFAULT_CONTEXT
    )
    reading_order = CompactCollectionProperty(
        "readingOrder", required=True, role=RWPMCollectionRolesRegistry.READING_ORDER
    )
    resources = CompactCollectionProperty(
        "resources", required=False, role=RWPMCollectionRolesRegistry.RESOURCES
    )
    toc = CompactCollectionProperty(
        "toc", required=False, role=RWPMCollectionRolesRegistry.TOC
    )



def first_or_default(collection, default=None):
    """Return first element of the specified collection or the default value if the collection is empty.

    :param collection: Collection
    :type collection: Iterable

    :param default: Default value
    :type default: Any
    """
    element = next(iter(collection), None)

    if element is None:
        element = default

    return element


def all_unique(_list):
    """Return a boolean value indicating whether the list contains only unique items.

    :return: Boolean value indicating whether the list contains only unique items
    :rtype: bool
    """
    if not isinstance(_list, list):
        raise ValueError("List expected")

    seen = set()

    return not any(element in seen or seen.add(element) for element in _list)


def is_string(value):
    """Return a boolean value indicating whether the value is a string or not.

    :return: Boolean value indicating whether the value is a string or not
    :rtype: bool
    """
    return isinstance(value, str)


def encode(value):
    """Encode the string value using UTF-8 encoding.

    :param value: Value to be encoded
    :type value: Any

    :return: Encoded string value or the initial value if it has a non-string type
    :rtype: Any
    """
    if isinstance(value, list):
        return f"list({first_or_default(value)}, ...)"

    return value if is_string(value) else value


def cast(value, klass):
    """Casts the value to the specified class.

    :param value: Value to be casted
    :type value: object

    :param klass: Target class
    :type klass: Type

    :return: Value casted to the specified class
    :rtype: object
    """
    if not isinstance(value, klass):
        raise ValueError(f"Value must be a transitive instance of {klass} class")

    value.__class__ = klass

    return value



from functools import partial

from multipledispatch import dispatch

from webpub_manifest_parser.core.ast import (
    Collection,
    CollectionList,
    CompactCollection,
    Link,
    LinkList,
    Manifestlike,
    Metadata,
)
from webpub_manifest_parser.core.semantic import (
    ManifestSemanticError,
    SemanticAnalyzer,
    SemanticAnalyzerError,
)
from webpub_manifest_parser.odl.ast import ODLFeed, ODLLicense, ODLPublication
from webpub_manifest_parser.odl.registry import ODLMediaTypesRegistry
from webpub_manifest_parser.opds2.ast import OPDS2FeedMetadata
from webpub_manifest_parser.opds2.registry import OPDS2LinkRelationsRegistry
from webpub_manifest_parser.utils import encode, first_or_default


class ODLPublicationSemanticError(SemanticAnalyzerError):
    """Base class for semantic errors related to ODL 2.x feeds."""

    def _format_message(
        self, node, node_property=None, message=None, inner_exception=None
    ):
        """Format the error message.

        :param node: ODL 2.x publication
        :type node: webpub_manifest_parser.odl.ast.ODLPublication

        :param node_property: AST node's property associated with the error
        :type node_property: Optional[webpub_manifest_parser.core.properties.Property]

        :param message: Parameterized string containing description of the error occurred
            and the placeholder for the feed's identifier
        :type message: Optional[str]

        :param inner_exception: (Optional) Inner exception
        :type inner_exception: Optional[Exception]
        """
        if not isinstance(node, ODLPublication):
            raise ValueError(f"Argument 'node' must be an instance of {ODLPublication}")

        if node.metadata:
            if node.metadata.title:
                message = message.format(node.metadata.title)
            elif node.metadata.identifier:
                message = message.format(node.metadata.identifier)

        return message


class ODLLicenseSemanticError(SemanticAnalyzerError):
    """Base class for semantic errors related to ODL 2.x feeds."""

    def _format_message(
        self, node, node_property=None, message=None, inner_exception=None
    ):
        """Format the error message.

        :param node: ODL 2.x license
        :type node: webpub_manifest_parser.odl.ast.ODLLicense

        :param node_property: AST node's property associated with the error
        :type node_property: Optional[webpub_manifest_parser.core.properties.Property]

        :param message: Parameterized string containing description of the error occurred
            and the placeholder for the feed's identifier
        :type message: Optional[str]

        :param inner_exception: (Optional) Inner exception
        :type inner_exception: Optional[Exception]
        """
        if not isinstance(node, ODLLicense):
            raise ValueError(
                f"Argument 'node' must be an instance of {ODLLicense} class"
            )

        if node.metadata:
            message = message.format(node.metadata.identifier)

        return message


ODL_FEED_MISSING_PUBLICATIONS_SUBCOLLECTION_ERROR = partial(
    ManifestSemanticError,
    message="ODL feed '{0}' does not contain required 'publications' subcollection",
)

ODL_FEED_CONTAINS_REDUNDANT_GROUPS_SUBCOLLECTIONS_ERROR = partial(
    ManifestSemanticError,
    message="ODL feed '{0}' contains redundant 'groups' subcollections",
)

ODL_FEED_CONTAINS_REDUNDANT_FACETS_SUBCOLLECTIONS_ERROR = partial(
    ManifestSemanticError,
    message="ODL feed '{0}' contains redundant 'facets' subcollections",
)

ODL_FEED_CONTAINS_REDUNDANT_NAVIGATION_SUBCOLLECTION_ERROR = partial(
    ManifestSemanticError,
    message="ODL feed '{0}' contains redundant 'navigation' subcollection",
)

ODL_PUBLICATION_MUST_CONTAIN_EITHER_LICENSES_OR_OA_ACQUISITION_LINK_ERROR = partial(
    ODLPublicationSemanticError,
    message="ODL publication '{0}' contains neither 'licenses' subcollection nor "
    "an Open-Access Acquisition Link (http://opds-spec.org/acquisition/open-access)",
)

ODL_LICENSE_MUST_CONTAIN_SELF_LINK_TO_LICENSE_INFO_DOCUMENT_ERROR = partial(
    ODLLicenseSemanticError,
    message="ODL license '{0}' does not contain a 'self' link to the License Info Document",
)

ODL_LICENSE_MUST_CONTAIN_CHECKOUT_LINK_TO_LICENSE_STATUS_DOCUMENT_ERROR = partial(
    ODLLicenseSemanticError,
    message="ODL license '{0}' does not contain a 'checkout' link to the License Status Document",
)


class ODLSemanticAnalyzer(SemanticAnalyzer):
    """ODL semantic analyzer."""

    @dispatch(Manifestlike)
    def visit(self, node):
        """Perform semantic analysis of the manifest node.

        :param node: Manifest-like node
        :type node: Manifestlike
        """
        super().visit(node)

        if not node.publications:
            with self._record_errors():
                raise ODL_FEED_MISSING_PUBLICATIONS_SUBCOLLECTION_ERROR(
                    node=node, node_property=ODLFeed.publications
                )

        if node.groups:
            with self._record_errors():
                raise ODL_FEED_CONTAINS_REDUNDANT_GROUPS_SUBCOLLECTIONS_ERROR(
                    node=node, node_property=ODLFeed.groups
                )

        if node.facets:
            with self._record_errors():
                raise ODL_FEED_CONTAINS_REDUNDANT_FACETS_SUBCOLLECTIONS_ERROR(
                    node=node, node_property=ODLFeed.facets
                )

        if node.navigation:
            with self._record_errors():
                raise ODL_FEED_CONTAINS_REDUNDANT_NAVIGATION_SUBCOLLECTION_ERROR(
                    node=node, node_property=ODLFeed.navigation
                )

        if node.links is not None:
            with self._record_errors():
                node.links.accept(self)
        if node.publications is not None:
            with self._record_errors():
                node.publications.accept(self)

    @dispatch(OPDS2FeedMetadata)
    def visit(self, node):
        """Perform semantic analysis of the feed's metadata.

        :param node: Feed's metadata
        :type node: OPDS2FeedMetadata
        """
        # super(ODLSemanticAnalyzer, self).visit(node)

    @dispatch(Metadata)
    def visit(self, node):
        """Perform semantic analysis of the feed's metadata.

        :param node: Feed's metadata
        :type node: Metadata
        """
        super().visit(node)

    @dispatch(ODLPublication)
    def visit(self, node):
        """Perform semantic analysis of the OPDS 2.0 publication.

        :param node: ODL 2.0 publication
        :type node: ODLPublication
        """
        self._logger.debug(f"Started processing {encode(node)}")

        if (not node.licenses or len(node.licenses) == 0) and (
            (not node.licenses or len(node.links) == 0)
            or not node.links.get_by_rel(OPDS2LinkRelationsRegistry.OPEN_ACCESS.key)
        ):
            with self._record_errors():
                raise ODL_PUBLICATION_MUST_CONTAIN_EITHER_LICENSES_OR_OA_ACQUISITION_LINK_ERROR(
                    node=node, node_property=None
                )
        elif node.licenses:
            node.licenses.accept(self)

        self._logger.debug(f"Finished processing {encode(node)}")

    @dispatch(LinkList)
    def visit(self, node):
        """Perform semantic analysis of the list of links.

        :param node: Manifest's metadata
        :type node: LinkList
        """
        super().visit(node)

    @dispatch(Link)
    def visit(self, node):
        """Perform semantic analysis of the link node.

        :param node: Link node
        :type node: Link
        """
        super().visit(node)

    @dispatch(CollectionList)
    def visit(self, node):
        """Perform semantic analysis of the list of sub-collections.

        :param node: CollectionList node
        :type node: CollectionList
        """
        super().visit(node)

    @dispatch(CompactCollection)
    def visit(self, node):
        """Perform semantic analysis of the compact collection node.

        :param node: Collection node
        :type node: CompactCollection
        """
        super().visit(node)

    @dispatch(Collection)
    def visit(self, node):
        """Perform semantic analysis of the collection node.

        :param node: Collection node
        :type node: Collection
        """
        super().visit(node)

    @dispatch(ODLLicense)
    def visit(self, node):
        """Perform semantic analysis of the ODL license node.

        :param node: ODLLicense node
        :type node: ODLLicense
        """
        self_link = (
            first_or_default(node.links.get_by_rel(OPDS2LinkRelationsRegistry.SELF.key))
            if node.links
            else None
        )

        if (
            not self_link
            or self_link.type != ODLMediaTypesRegistry.ODL_LICENSE_INFO_DOCUMENT.key
        ):
            with self._record_errors():
                raise ODL_LICENSE_MUST_CONTAIN_SELF_LINK_TO_LICENSE_INFO_DOCUMENT_ERROR(
                    node=node, node_property=None
                )

        borrow_link = (
            first_or_default(
                node.links.get_by_rel(OPDS2LinkRelationsRegistry.BORROW.key)
            )
            if node.links
            else None
        )

        if (
            not borrow_link
            or borrow_link.type != ODLMediaTypesRegistry.ODL_LICENSE_STATUS_DOCUMENT.key
        ):
            with self._record_errors():
                raise ODL_LICENSE_MUST_CONTAIN_CHECKOUT_LINK_TO_LICENSE_STATUS_DOCUMENT_ERROR(
                    node=node, node_property=None
                )



from webpub_manifest_parser.core.ast import LinkProperties, Metadata, Node
from webpub_manifest_parser.core.parsers import ArrayParser, EnumParser
from webpub_manifest_parser.core.properties import (
    EnumProperty,
    IntegerProperty,
    Property,
    StringProperty,
    TypeProperty,
    URIProperty,
)


class EPUBPresentationHints(Node):
    """EPUB presentation hints."""

    layout = EnumProperty("layout", required=False, items=["fixed", "reflowable"])


class EPUBMetadata(Metadata):
    """EPUB metadata."""

    presentation = TypeProperty(
        "presentation", required=False, nested_type=EPUBPresentationHints
    )


class EPUBEncryptionSettings(Node):
    """EPUB encryption settings."""

    algorithm = URIProperty("algorithm", required=True)
    compression = StringProperty("compression", required=False)
    original_length = IntegerProperty("originalLength", required=False)
    profile = URIProperty("profile", required=False)
    scheme = URIProperty("scheme", required=False)


class EPUBLinkProperties(LinkProperties):
    """EPUB link properties."""

    contains = Property(
        "contains",
        required=False,
        parser=ArrayParser(
            item_parser=EnumParser(
                ["mathml", "onix", "remote-resources", "js", "svg", "xmp"]
            ),
            unique_items=True,
        ),
    )
    layout = EnumProperty("layout", False, ["fixed", "reflowable"])
    encryption_settings = TypeProperty("encrypted", False, EPUBEncryptionSettings)



class BaseError(Exception):
    """Base class for all errors."""

    def __init__(self, message=None, inner_exception=None):
        """Initialize a new instance of BaseError class.

        :param message: String containing description of the error occurred
        :type message: Optional[str]

        :param inner_exception: (Optional) Inner exception
        :type inner_exception: Optional[Exception]
        """
        if inner_exception and not message:
            message = inner_exception.error_message

        super().__init__(message)

        self._inner_exception = inner_exception

    @property
    def inner_exception(self):
        """Return the inner exception.

        :return: Inner exception
        :rtype: Exception
        """
        return self._inner_exception

    @property
    def error_message(self):
        """Return the description of the error.

        :return: Description of the error
        :rtype: str
        """
        return str(self)

    def __eq__(self, other):
        """Compare two BaseError objects.

        :param other: BaseError object
        :type other: BaseError

        :return: Boolean value indicating whether two items are equal
        :rtype: bool
        """
        if not isinstance(other, BaseError):
            return False

        return str(self) == str(other)

    def __repr__(self):
        """Return error's string representation.

        :return: Error's string representation
        :rtype: str
        """
        return "<BaseError(message={}, inner_exception={})>".format(
            self.error_message, self.inner_exception
        )



from webpub_manifest_parser.core.registry import CollectionRole, MediaType
from webpub_manifest_parser.opds2.registry import (
    OPDS2CollectionRolesRegistry,
    OPDS2MediaTypesRegistry,
)


class ODLMediaTypesRegistry(OPDS2MediaTypesRegistry):
    """Registry containing ODL 2.0 media types."""

    ODL_LICENSE_INFO_DOCUMENT = MediaType(key="application/vnd.odl.info+json")
    ODL_LICENSE_STATUS_DOCUMENT = MediaType(
        key="application/vnd.readium.license.status.v1.0+json"
    )


class ODLCollectionRolesRegistry(OPDS2CollectionRolesRegistry):
    """Registry containing ODPS 2.0 collection roles."""

    LICENSES = CollectionRole(key="licenses", compact=True, required=False)

    ODL_ROLES = [LICENSES] + OPDS2CollectionRolesRegistry.OPDS_2_0_ROLES

    def __init__(self):
        """Initialize a new instance of OPDS2CollectionRolesRegistry class."""
        super().__init__()

        self._add_items(self.ODL_ROLES)



import logging
import re
from abc import ABCMeta, abstractmethod
from pydoc import locate

import jsonschema
from dateutil import parser as datetime_parser
from jsonschema import FormatError
from uritemplate import URITemplate

from webpub_manifest_parser.errors import BaseError
from webpub_manifest_parser.utils import encode, is_string


class ValueParserError(BaseError):
    """Base class for all errors raised by value parsers."""

    def __init__(self, value, message, inner_exception=None):
        """Initialize a new instance of ValueParsingError class.

        :param value: Value associated with this error
        :type value: Any

        :param message: String containing description of the error occurred
        :type message: basestring

        :param inner_exception: (Optional) Inner exception
        :type inner_exception: Optional[Exception]
        """
        super().__init__(message, inner_exception)

        self._value = value

    @property
    def value(self):
        """Return the value associated with this error.

        :return: Value associated with this error
        :rtype: Any
        """
        return self._value


class ValueParser(metaclass=ABCMeta):
    """Base parser class."""

    @abstractmethod
    def parse(self, value):
        """Parse the value, raise ParsingError if the value is not correct, otherwise return the processed value.

        :param value: Value to validate
        :type value: Any

        :return: Processed value (for example, datetime object in the case of a value containing a date&time string)
        :rtype: Any

        :raise: ValidationError
        """
        raise NotImplementedError()


class AnyOfParser(ValueParser):
    """Parser making sure that at least one of the inner parsers succeed."""

    def __init__(self, inner_parsers):
        """Initialize a new instance of AnyOfParser class.

        :param inner_parsers: List of composed parsers
        :type inner_parsers: List[ValueParser]
        """
        self._inner_parsers = inner_parsers
        self._logger = logging.getLogger(__name__)

    @property
    def inner_parsers(self):
        """Return the inner parsers.

        :return: Inner parsers
        :rtype: List[ValueParser]
        """
        return self._inner_parsers

    def parse(self, value):
        """Make sure that at least one of the inner parsers succeed, otherwise raise the first validation error.

        :param value: Value
        :type value: Any

        :return: First valid value
        :rtype: Any

        :raise: ValidationError
        """
        first_validation_error = None

        for parser in self._inner_parsers:
            self._logger.debug(f"Running {parser} parser")

            try:
                result = parser.parse(value)

                self._logger.debug(f"Parser {parser} succeeded: {encode(result)}")

                return result
            except ValueParserError as error:
                self._logger.debug(f"Parser {encode(parser)} failed")

                if first_validation_error is None:
                    first_validation_error = error

        self._logger.debug("All parsers failed")

        raise first_validation_error


class NumericParser(ValueParser, metaclass=ABCMeta):
    """Numeric parser."""

    def __init__(
        self, minimum=None, exclusive_minimum=None, maximum=None, exclusive_maximum=None
    ):
        """Initialize a new instance of NumericParser class.

        :param minimum: Minimum value
        :type minimum: Numeric

        :param exclusive_minimum: Exclusive minimum value
        :type exclusive_minimum: Numeric

        :param maximum: Maximum value
        :type maximum: Numeric

        :param exclusive_maximum: Exclusive maximum value
        :type exclusive_maximum: Numeric
        """
        self._minimum = minimum
        self._exclusive_minimum = exclusive_minimum
        self._maximum = maximum
        self._exclusive_maximum = exclusive_maximum

    @abstractmethod
    def _parse(self, value):
        """Parse a numeric string into a Python numeric object (int or float).

        :param value: Value
        :type value: Any

        :return: Parsed numeric value
        :rtype: Numeric

        :raise: ValidationError
        """
        raise NotImplementedError()

    def parse(self, value):
        """Parse a numeric value.

        :param value: Value
        :type value: Any

        :return: Parsed numeric value
        :rtype: Numeric

        :raise: ValidationError
        """
        value = self._parse(value)

        if self._minimum is not None and value < self._minimum:
            raise ValueParserError(
                value,
                f"Value {value} is less than the minimum ({self._minimum})",
            )
        if self._exclusive_minimum is not None and value <= self._exclusive_minimum:
            raise ValueParserError(
                value,
                "Value {} is less or equal than the exclusive minimum ({})".format(
                    value, self._exclusive_minimum
                ),
            )
        if self._maximum is not None and value > self._maximum:
            raise ValueParserError(
                value,
                "Value {} is greater than the maximum ({})".format(
                    value, self._maximum
                ),
            )
        if self._exclusive_maximum is not None and value >= self._exclusive_maximum:
            raise ValueParserError(
                value,
                "Value {} is greater or equal than the exclusive maximum ({})".format(
                    value, self._exclusive_maximum
                ),
            )

        return value


class IntegerParser(NumericParser):
    """Integer parser."""

    def _parse(self, value):
        """Parse an integer value.

        :param value: Value
        :type value: Any

        :return: Parsed integer
        :rtype: int

        :raise: ValidationError
        """
        try:
            return int(value)
        except ValueError as error:
            raise ValueParserError(value, str(error), error)


class NumberParser(NumericParser):
    """Number parser."""

    def _parse(self, value):
        """Parse a float number.

        :param value: Value
        :type value: Any

        :return: Parsed float number
        :rtype: int

        :raise: ValidationError
        """
        try:
            return float(value)
        except ValueError as error:
            raise ValueParserError(value, str(error), error)


class BooleanParser(ValueParser):
    """Boolean parser."""

    def parse(self, value):
        """Parse a boolean value.

        :param value: Value
        :type value: Any

        :return: Parsed boolean value
        :rtype: int

        :raise: ValidationError
        """
        if isinstance(value, bool):
            return value

        if is_string(value):
            if value == "false":
                return False

            if value == "true":
                return True

        raise ValueParserError(value, f"Value '{encode(value)}' must be boolean")


class StringParser(ValueParser):
    """String parser."""

    def parse(self, value):
        """Parse a string value.

        :param value: Value
        :type value: Any

        :return: Parsed string value
        :rtype: str

        :raise: ValidationError
        """
        if not is_string(value):
            raise ValueParserError(value, f"Value '{encode(value)}' must be a string")

        return value


class StringPatternParser(StringParser):
    """String parser using a regular expression pattern."""

    def __init__(self, pattern):
        """Initialize a new instance of StringPatternParser class.

        :param pattern: Regular expression which string's value must conform to
        :type pattern: RegEx
        """
        if not is_string(pattern):
            raise ValueError("Argument 'pattern' must be a string")

        self._pattern = pattern
        self._regex = re.compile(pattern) if pattern is not None else None

    def parse(self, value):
        """Parse a string value using the specified regular expression.

        :param value: Value
        :type value: Any

        :return: Parsed string value
        :rtype: int

        :raise: ValidationError
        """
        value = super().parse(value)

        if not self._regex.match(value):
            raise ValueParserError(
                value,
                "String value '{}' does not match regular expression {}".format(
                    encode(value), self._pattern
                ),
            )

        return value


class EnumParser(StringParser):
    """Enum parser."""

    def __init__(self, items):
        """Initialize a new instance of EnumParser class.

        :param items: Enumeration items
        :type items: List[str]
        """
        self._items = items

    def parse(self, value):
        """Make sure that the value is a part of the enumeration and return it back.

        :param value: Value
        :type value: Any

        :return: Parsed string value
        :rtype: int

        :raise: ValidationError
        """
        value = super().parse(value)

        if value not in self._items:
            raise ValueParserError(
                value,
                f"Value '{encode(value)}' is not among {self._items}",
            )

        return value


class FormatChecker(StringParser, metaclass=ABCMeta):
    """Base class for all parsers using jsonschema FormatChecker."""

    def __init__(self, json_schema_format):
        """Initialize a new instance of FormatParser class.

        :param json_schema_format: One of the jsonschema allowed string formats (color, date, date-time, etc.)
        :type json_schema_format: str
        """
        self._json_schema_format = json_schema_format
        self._format_checker = jsonschema.FormatChecker([json_schema_format])

    def _validate(self, value):
        """Check the value's format.

        :param value: Value
        :type value: Any

        :raise: ValidationError
        """
        try:
            self._format_checker.check(value, self._json_schema_format)
        except FormatError as error:
            raise ValueParserError(value, str(error), error)

    def _parse(self, value):
        """Parse the value into an appropriate Python type.

        :param value: Value
        :type value: Any

        :return: Parsed value
        :rtype: Any

        :raise: ValidationError
        """
        return value

    def parse(self, value):
        """Parse the value according to the specified format.

        :param value: Value
        :type value: Any

        :return: Parsed value
        :rtype: Any

        :raise: ValidationError
        """
        self._validate(value)
        result = self._parse(value)

        return result


class URIParser(FormatChecker):
    """URI parser."""

    def __init__(self):
        """Initialize a new instance of URIParser class."""
        super().__init__("uri")

    def _parse(self, value):
        """Parse the URI into a Python dictionary containing URI's subcomponents.

        :param value: Value
        :type value: Any

        :return: URI
        :rtype: str
        """
        return value


class URITemplateParser(StringParser):
    """URI template parser."""

    def parse(self, value):
        """Parse the URI into a Python dictionary containing URI's subcomponents.

        :param value: Value
        :type value: Any

        :return: URI template
        :rtype: URITemplate
        """
        URITemplate(value)

        return value


class URIReferenceParser(FormatChecker):
    """URI-reference parser."""

    def __init__(self):
        """Initialize a new instance of URIReferenceParser class."""
        super().__init__("uri-reference")

    def _parse(self, value):
        """Parse the URI-reference into a Python dictionary containing URI's subcomponents.

        :param value: Value
        :type value: Any

        :return: URI-reference
        :rtype: str
        """
        return value


class DateParser(StringParser):
    """Date parser."""

    def _raise_error(self, value, inner_exception=None):
        raise ValueParserError(
            value,
            f"Value '{encode(value)}' is not a correct date",
            inner_exception,
        )

    def parse(self, value):
        """Parse a date & time string into datetime object.

        :param value: Value
        :type value: Any

        :return: Parsed date object
        :rtype: datetime.datetime
        """
        value = super().parse(value)
        result = None

        try:
            result = datetime_parser.isoparse(value)
        except Exception as exception:
            self._raise_error(value, exception)

        if result and (
            result.hour
            or result.minute
            or result.second
            or result.microsecond
            or result.tzinfo
        ):
            self._raise_error(value)

        return result


class DateTimeParser(StringParser):
    """Date & time parser."""

    def parse(self, value):
        """Parse a date & time string into datetime object.

        :param value: Value
        :type value: Any

        :return: Parsed date & time object
        :rtype: datetime.datetime
        """
        value = super().parse(value)

        try:
            return datetime_parser.isoparse(value)
        except Exception as exception:
            raise ValueParserError(
                value,
                "Value '{}' is not a correct date & time value: "
                "it does not comply with ISO 8601 date & time formatting rules".format(
                    encode(value)
                ),
                exception,
            )


class ArrayParser(ValueParser):
    """Array parser."""

    def __init__(self, item_parser, unique_items=False):
        """Initialize a new instance of ArrayParser class.

        :param item_parser: Parser used for parsing array items
        :type item_parser: ValueParser

        :param unique_items: Boolean value indicating whether array must contain only unique items
        :type unique_items: bool
        """
        self._item_parser = item_parser
        self._unique_items = unique_items
        self._logger = logging.getLogger(__name__)

    @property
    def item_parser(self):
        """Return the parser used for parsing array items.

        :return: Parser used for parsing array items
        :rtype: ValueParser
        """
        return self._item_parser

    def parse(self, value):
        """Parse the value into a list of parsed values.

        :param value: Value
        :type value: Any

        :return: List consisting of parsed items
        :rtype: List

        :raise: ValidationError
        """
        if not isinstance(value, list):
            raise ValueParserError(value, f"Value '{encode(value)}' must be a list")

        result = []
        seen = set()

        for item in value:
            item = self._item_parser.parse(item)

            if self._unique_items and item in seen:
                raise ValueParserError(value, f"Item '{encode(item)}' is not unique")

            result.append(item)
            seen.add(item)

        return list(result)


class ObjectParser(ValueParser):
    """Object parser."""

    def __init__(self, properties_parser, properties_pattern=None):
        """Initialize a new instance of ObjectParser class.

        :param properties_parser: Properties parser
        :type properties_parser: ValueParser

        :param properties_pattern: Properties regex pattern
        :type properties_pattern: Optional[str]
        """
        self._properties_parser = properties_parser
        self._properties_regex = (
            re.compile(properties_pattern) if properties_pattern is not None else None
        )

    def parse(self, value):
        """Parse a JSON object into a Python dictionary.

        :param value: Value
        :type value: Any

        :return: Python dictionary containing parsed items
        :rtype: Dict

        :raise: ValidationError
        """
        if not isinstance(value, dict):
            raise ValueParserError(value, "Value must be a dictionary")

        result = {}

        for key, item in list(value.items()):
            if not isinstance(key, str):
                raise ValueParserError(value, f"Key '{encode(key)}' must be a string")

            if self._properties_regex and not self._properties_regex.match(key):
                raise ValueParserError(
                    value,
                    "Key '{}' does not match the pattern '{}'".format(
                        encode(key), self._properties_regex
                    ),
                )

            item = self._properties_parser.parse(item)

            result[key] = item

        return result


class LocalizableStringParser(ObjectParser):
    """Parser for localizable strings represented as a dictionary with keys equal to language codes."""

    LANGUAGE_PATTERN = "^((?P<grandfathered>(en-GB-oed|i-ami|i-bnn|i-default|i-enochian|i-hak|i-klingon|i-lux|i-mingo|i-navajo|i-pwn|i-tao|i-tay|i-tsu|sgn-BE-FR|sgn-BE-NL|sgn-CH-DE)|(art-lojban|cel-gaulish|no-bok|no-nyn|zh-guoyu|zh-hakka|zh-min|zh-min-nan|zh-xiang))|((?P<language>([A-Za-z]{2,3}(-(?P<extlang>[A-Za-z]{3}(-[A-Za-z]{3}){0,2}))?)|[A-Za-z]{4}|[A-Za-z]{5,8})(-(?P<script>[A-Za-z]{4}))?(-(?P<region>[A-Za-z]{2}|[0-9]{3}))?(-(?P<variant>[A-Za-z0-9]{5,8}|[0-9][A-Za-z0-9]{3}))*(-(?P<extension>[0-9A-WY-Za-wy-z](-[A-Za-z0-9]{2,8})+))*(-(?P<privateUse>x(-[A-Za-z0-9]{1,8})+))?)|(?P<privateUse2>x(-[A-Za-z0-9]{1,8})+))$"

    def __init__(self):
        """Initialize a new instance of LocalizableStringValidator class."""
        super().__init__(StringParser(), self.LANGUAGE_PATTERN)


class TypeParser(ValueParser):
    """Parser used for checking type information and as an indication of nested types."""

    def __init__(self, _type):
        """Initialize a new instance of TypeParser class.

        :param _type: Type
        :type _type: Union[Type, str]
        """
        self._type = _type

    @property
    def type(self):
        """Return the type.

        "return: Type
        :rtype: Type
        """
        if is_string(self._type):
            self._type = locate(self._type)

            if self._type is None:
                raise ValueError(f"Unknown type {self._type}")

        return self._type

    def parse(self, value):
        """Check that the value has the correct type.

        :param value: Value
        :type value: Any

        :return: Value
        :rtype: Any

        :raise: ValidationError
        """
        if not isinstance(value, self._type):
            raise ValueParserError(
                value,
                "Value '{}' must be an instance of '{}'".format(
                    encode(value), self._type
                ),
            )

        return value


def find_parser(parent_parser, child_parser_type):
    """Find a child parser with a specified type in the parent parser.

    This function is used for finding nested types.

    :param parent_parser: Parent parser
    :type parent_parser: ValueParser

    :param child_parser_type: Child parser's type
    :type child_parser_type: Type

    :return: List of 2-tuples (parent parser, child parser)
    :rtype: List[Optional[ValueParser], ValueParser]
    """
    candidates = []

    def _find_parser(_parent_parser, _current_parser):
        if isinstance(_current_parser, child_parser_type):
            candidates.append((_parent_parser, _current_parser))
        elif isinstance(_current_parser, AnyOfParser):
            for inner_parser in _current_parser.inner_parsers:
                _find_parser(_parent_parser, inner_parser)
        elif isinstance(_current_parser, ArrayParser):
            _find_parser(_current_parser, _current_parser.item_parser)

    _find_parser(None, parent_parser)

    return candidates



from abc import ABCMeta
from contextlib import contextmanager

from webpub_manifest_parser.core.properties import BaseArrayProperty, PropertiesGrouping
from webpub_manifest_parser.errors import BaseError


class BaseAnalyzerError(BaseError):
    """Exception raised in the case of any (syntax, semantic) errors thrown during parsing."""

    def __init__(self, node, node_property, message=None, inner_exception=None):
        """Initialize a new instance of BaseSemanticError class.

        :param node: AST node where the error was found
        :type node: webpub_manifest_parser.core.ast.Node

        :param node_property: AST node's property associated with the error
        :type node_property: Optional[webpub_manifest_parser.core.properties.Property]

        :param message: String containing description of the error occurred
        :type message: Optional[str]

        :param inner_exception: (Optional) Inner exception
        :type inner_exception: Optional[Exception]
        """
        super().__init__(message, inner_exception)

        self._node = node
        self._node_property = node_property

    @property
    def node(self):
        """Return the AST node where the error was found.

        :return: Node where the error was found
        :rtype: webpub_manifest_parser.core.ast.Node
        """
        return self._node

    @property
    def node_property(self):
        """Return the AST node's property associated with the error.

        :return: AST node's property associated with the error
        :rtype: webpub_manifest_parser.core.properties.Property
        """
        return self._node_property


class AnalyzerContext:
    """Class containing the current analyzer's context."""

    def __init__(self):
        """Initialize a new instance of AnalyzerContext class."""
        self._errors = []

    @property
    def errors(self):
        """Return the list of errors.

        :return: List of errors
        :rtype: List[BaseAnalyzerError]
        """
        return self._errors

    def reset(self):
        """Reset the current context."""
        self._errors = []


class BaseAnalyzer(metaclass=ABCMeta):
    """Base class for all analyzers (syntax and semantic)."""

    def __init__(self):
        """Initialize a new instance of BaseAnalyzer class."""
        self._context = AnalyzerContext()

    @property
    def context(self):
        """Return the current analyzer's context.

        :return: Current analyzer's context
        :rtype: AnalyzerContext
        """
        return self._context

    @contextmanager
    def _record_errors(self):
        """Record semantic errors in the current context.

        NOTE: This is a very naive implementation of error recovery.
        """
        try:
            yield
        except BaseAnalyzerError as error:
            self.context.errors.append(error)


class NodeFinder:
    """Used for traversing the AST."""

    def _find_parent_or_self(
        self, current_node, target_node, target_parent_class, parent_nodes=None
    ):
        """In the AST defined by `root` for the given `target_node` find a parent node with `target_parent_class`.

        :param current_node: The root of the current sub-tree
        :type current_node: webpub_manifest_parser.core.ast.Node

        :param target_node: Target AST node which parent we want to find
        :type target_node: webpub_manifest_parser.core.ast.Node

        :param target_parent_class: Class of the parent node
        :type target_parent_class: Type

        :param parent_nodes: List of the parent nodes
        :type parent_nodes: list

        :return: Parent of the `target_node` with `target_parent_class`
        :rtype target_node: webpub_manifest_parser.core.ast.Node
        """
        if current_node == target_node:
            if isinstance(current_node, target_parent_class):
                return current_node

            while parent_nodes:
                parent_node = parent_nodes.pop()

                if isinstance(parent_node, target_parent_class):
                    return parent_node

        parent_nodes = parent_nodes if parent_nodes else []
        parent_nodes.append(current_node)

        ast_object_properties = PropertiesGrouping.get_class_properties(
            current_node.__class__
        )

        for object_property_name, object_property in ast_object_properties:
            if isinstance(object_property, BaseArrayProperty):
                child_nodes = getattr(current_node, object_property_name)

                if not child_nodes:
                    continue

                for child_node in child_nodes:
                    parent = self._find_parent_or_self(
                        child_node, target_node, target_parent_class, parent_nodes
                    )

                    if parent:
                        return parent
            else:
                child_node = getattr(current_node, object_property_name)

                if not child_node:
                    continue

                parent = self._find_parent_or_self(
                    child_node, target_node, target_parent_class, parent_nodes
                )

                if parent:
                    return parent

        return None

    def find_parent_or_self(self, root, target_node, target_parent_class):
        """In the AST defined by `root` for the given `target_node` find a parent node with `target_parent_class`.

        :param root: Root node of the AST
        :type root: webpub_manifest_parser.core.ast.Node

        :param target_node: Target AST node whose parent we want to find
        :type target_node: webpub_manifest_parser.core.ast.Node

        :param target_parent_class: Class of the parent node
        :type target_parent_class: Type

        :return: Parent of the `target_node` with `target_parent_class`
        :rtype target_node: webpub_manifest_parser.core.ast.Node
        """
        return self._find_parent_or_self(root, target_node, target_parent_class)



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import TypeVar

from webpub_manifest_parser.core.parsers import (
    AnyOfParser,
    ArrayParser,
    StringParser,
    TypeParser,
)
from webpub_manifest_parser.core.properties import (
    ArrayOfStringsProperty,
    ArrayOfURIsProperty,
    ArrayProperty,
    BaseArrayProperty,
    BooleanProperty,
    DateOrTimeProperty,
    DateTimeProperty,
    EnumProperty,
    IntegerProperty,
    ListOfLanguagesProperty,
    LocalizableStringProperty,
    NumberProperty,
    PropertiesGrouping,
    Property,
    StringProperty,
    TypeProperty,
    URIProperty,
    URITemplateProperty,
)
from webpub_manifest_parser.core.registry import CollectionRole
from webpub_manifest_parser.utils import encode

T = TypeVar("T")


def _safe_tuple(value: list[T] | T) -> tuple[T] | T:
    """Convert a value to a tuple if it is a list only"""
    return tuple(value) if isinstance(value, list) else value


class Visitor(metaclass=ABCMeta):
    """Interface for visitors walking through abstract syntax trees (AST)."""

    @abstractmethod
    def visit(self, node):
        """Process the specified node.

        :param node: AST node
        :type node: Node
        """
        raise NotImplementedError()


class Visitable(metaclass=ABCMeta):
    """Interface for objects walkable by AST visitors."""

    @abstractmethod
    def accept(self, visitor):
        """Accept  the specified visitor.

        :param visitor: Visitor object
        :type visitor: Visitor
        """
        raise NotImplementedError()


class Extendable(metaclass=ABCMeta):
    """Abstract class adding ability to extend classes.

    For example, RWPM link properties can be extended by EPUB link properties and OPDS 2.0 link properties.
    """

    extensions = None

    @classmethod
    def get_extension(cls):
        """Return a new class having all extensions as mixins.

        :return: New class containing extensions as mixins.
        :rtype: Type
        """
        if not cls.extensions:
            return cls

        if len(cls.extensions) == 1:
            return cls.extensions[0]

        class_names = [cls.__name__] + [
            extension.__name__ for extension in cls.extensions
        ]
        extended_class_name = "_".join(class_names)
        extended_class = type(extended_class_name, tuple(cls.extensions), {})

        return extended_class


class Node(PropertiesGrouping, Visitable, Extendable):
    """Base class for all AST nodes."""

    def accept(self, visitor):
        """Accept the specified visitor.

        :param visitor: Visitor object
        :type visitor: Visitor
        """
        visitor.visit(self)


class LinkProperties(Node):
    """Link properties."""

    clipped = BooleanProperty("clipped", required=False)
    fit = EnumProperty(
        "fit", required=False, items=["contain", "cover", "width", "height"]
    )
    orientation = EnumProperty(
        "orientation", required=False, items=["auto", "landscape", "portrait"]
    )
    page = EnumProperty("page", required=False, items=["left", "right", "center"])
    spread = EnumProperty(
        "spread", required=False, items=["auto", "both", "none", "landscape"]
    )

    def __hash__(self):
        """Calculate the hash.

        :return: Hash
        :rtype: int
        """
        return hash((self.clipped, self.orientation, self.page, self.spread))

    def __repr__(self):
        """Return a string representation of the object.

        :return: String representation
        :rtype: str
        """
        return "<LinkProperties(clipped={}, fit={}, orientation={}, page={}, spread={})>".format(
            self.clipped, self.fit, self.orientation, self.page, self.spread
        )


class Link(Node):
    """Link to another resource."""

    href = URITemplateProperty("href", required=True)
    templated = BooleanProperty("templated", required=False)
    type = StringProperty("type", required=False)
    title = StringProperty("title", required=False)
    rels = ArrayOfStringsProperty("rel", required=False)
    properties = TypeProperty("properties", required=False, nested_type=LinkProperties)
    height = IntegerProperty("height", required=False, exclusive_minimum=0)
    width = IntegerProperty("width", required=False, exclusive_minimum=0)
    bitrate = NumberProperty("bitrate", required=False, exclusive_minimum=0)
    duration = NumberProperty("duration", required=False, exclusive_minimum=0)
    languages = ListOfLanguagesProperty("language", required=False)
    alternates = ArrayProperty(
        "alternate",
        required=False,
        item_parser=TypeParser("webpub_manifest_parser.core.ast.Link"),
    )
    children = ArrayProperty(
        "children",
        required=False,
        item_parser=TypeParser("webpub_manifest_parser.core.ast.Link"),
    )

    def __init__(
        self,
        href=None,
        templated=None,
        _type=None,
        title=None,
        rels=None,
        properties=None,
        height=None,
        width=None,
        duration=None,
        bitrate=None,
        languages=None,
        alternates=None,
        children=None,
    ):
        """Initialize a new instance of Link class.

        :param href: Link's URL
        :type href: str

        :param templated: Boolean value indicating whether href is a URI template
        :type templated: bool

        :param _type: Media type of the linked resource
        :type _type: Union[str, MediaType]

        :param title: Title of the linked resource
        :type title: str

        :param rels: Relation between the resource and its containing collection
        :type rels: List[registry.LinkRelation]

        :param properties: Relation between the resource and its containing collection
        :type properties: object

        :param height: Height of the linked resource in pixels
        :type height: int

        :param width: Width of the linked resource in pixels
        :type width: int

        :param duration: Duration of the linked resource in seconds
        :type duration: float

        :param bitrate: Bit rate of the linked resource in kilobits per second
        :type bitrate: float

        :param languages: Expected languages of the linked resource
        :type languages: List[str]

        :param alternates: Alternate resources for the linked resource
        :type alternates: List[Link]

        :param children: Resources that are children of the linked resource, in the context of a given collection role
        :type children: List[Link]
        """
        super().__init__()

        self.href = href
        self.templated = templated
        self.type = _type
        self.title = title
        self.rels = rels
        self.properties = properties
        self.height = height
        self.width = width
        self.duration = duration
        self.bitrate = bitrate
        self.languages = languages
        self.alternates = alternates
        self.children = children

    def __eq__(self, other):
        """Compare two Link objects.

        :param other: Link object
        :type other: Link

        :return: Boolean value indicating whether two items are equal
        :rtype: bool
        """
        if not isinstance(other, Link):
            return False

        return (
            self.href == other.href
            and self.templated == other.templated
            and self.type == other.type
            and self.title == other.title
            and self.rels == other.rels
            and self.properties == other.properties
            and self.height == other.height
            and self.width == other.width
            and self.duration == other.duration
            and self.bitrate == other.bitrate
            and self.languages == other.languages
            and self.alternates == other.alternates
            and self.children == other.children
        )

    def __hash__(self):
        """Calculate the hash.

        :return: Hash
        :rtype: int
        """
        return hash(
            (
                self.href,
                self.templated,
                self.type,
                self.title,
                _safe_tuple(self.rels),
                self.properties,
                self.height,
                self.width,
                self.duration,
                self.bitrate,
                _safe_tuple(self.languages),
                _safe_tuple(self.alternates),
                _safe_tuple(self.children),
            )
        )

    def __repr__(self):
        """Return a string representation of the object.

        :return: String representation
        :rtype: str
        """
        return (
            "<Link("
            "href={}, "
            "templated={}, "
            "type={}, "
            "title={}, "
            "rels={}, "
            "properties={}, "
            "height={}, "
            "width={}, "
            "duration={}, "
            "bitrate={}, "
            "languages={}, "
            "alternates={}, "
            "children={}".format(
                self.href,
                self.templated,
                self.type,
                self.title,
                self.rels,
                self.properties,
                self.height,
                self.width,
                self.duration,
                self.bitrate,
                self.languages,
                self.alternates,
                self.children,
            )
        )


class LinkList(Node, list):
    """List of links."""

    def __init__(self, items=None):
        """Initialize a new instance of LinksList class.

        :param items: (Optional) Items to be added to the list
        :type items: Optional[List]
        """
        super().__init__()

        if items is not None:
            if not isinstance(items, list):
                raise ValueError(f"Argument 'items' must be an instance of {list}")

            self.extend(items)

    def __hash__(self):
        """Calculate the hash.

        :return: Hash
        :rtype: int
        """
        return hash(tuple(self))

    def get_by_rel(self, rel):
        """Return links with the specific relation.

        :param rel: Link's relation
        :type rel: str

        :return: Links with the specified relation
        :rtype: List[Link]
        """
        return [link for link in self if link.rels and rel in link.rels]

    def get_by_href(self, href):
        """Return links with the specific URL.

        :param href: Link's URL
        :type href: str

        :return: Links with the specified relation
        :rtype: List[Link]
        """
        return [link for link in self if href == link.href]


class ArrayOfLinksProperty(BaseArrayProperty):
    """Property allowing to contain only unique links."""

    def __init__(self, key, required):
        """Initialize a new instance of ArrayOfLinksProperty class.

        :param key: Property's key
        :type key: str

        :param required: Boolean value indicating whether the property is required or not
        :type required: bool
        """
        super().__init__(key, required, ArrayParser(TypeParser(Link), True), LinkList)


class Contributor(Node):
    """Contributor object."""

    name = LocalizableStringProperty("name", required=True)
    identifier = URIProperty("identifier", required=False)
    sort_as = StringProperty("sortAs", required=False)
    roles = ArrayOfStringsProperty("role", required=False)
    position = NumberProperty("position", required=False)
    links = ArrayOfLinksProperty("links", required=False)

    def __init__(
        self,
        name=None,
        identifier=None,
        sort_as=None,
        roles=None,
        position=None,
        links=None,
    ):
        """Initialize a new instance of Contributor class."""
        super().__init__()

        self.name = name
        self.identifier = identifier
        self.sort_as = sort_as
        self.roles = roles
        self.position = position
        self.links = links

    def __eq__(self, other):
        """Compare two Contributor objects.

        :param other: Contributor object
        :type other: Contributor

        :return: Boolean value indicating whether two items are equal
        :rtype: bool
        """
        if not isinstance(other, Contributor):
            return False

        return (
            self.name == other.name
            and self.identifier == other.identifier
            and self.sort_as == other.sort_as
            and self.roles == other.roles
            and self.position == other.position
            and self.links == other.links
        )

    def __hash__(self):
        """Calculate the hash.

        :return: Hash
        :rtype: int
        """
        return hash(
            (
                self.name,
                self.identifier,
                self.sort_as,
                tuple(self.roles) if self.roles else tuple(),
                self.position,
                tuple(self.links) if self.links else tuple(),
            )
        )

    def __repr__(self):
        """Return a string representation of the object.

        :return: String representation
        :rtype: str
        """
        return "<Contributor(name={}, identifier={}, sort_as={}, roles={}, position={}, links={})>".format(
            encode(self.name),
            self.identifier,
            self.sort_as,
            self.roles,
            self.position,
            self.links,
        )


class ArrayOfContributorsProperty(BaseArrayProperty):
    """Property containing information about contributors.

    For example:
        - "Herman Melville"
        - {
            name: "Herman Melville"
          }
        - [
            "Herman Melville",
            "Mark Twain"
          ]
        - [
            {
                name: "Herman Melville"
            },
            {
                name: "Mark Twain"
            }
          ]
    """

    PARSER = AnyOfParser(
        [
            StringParser(),
            ArrayParser(AnyOfParser([StringParser(), TypeParser(Contributor)])),
            TypeParser(Contributor),
        ]
    )

    def __init__(self, key, required):
        """Initialize a new instance of ArrayOfContributorsProperty class.

        :param key: Property's key
        :type key: str

        :param required: Boolean value indicating whether the property is required or not
        :type required: bool
        """
        super().__init__(key, required, self.PARSER, list, [])


class Subject(Node, PropertiesGrouping):
    """Subject object."""

    name = LocalizableStringProperty("name", required=True)
    sort_as = StringProperty("sortAs", required=False)
    code = StringProperty("code", required=False)
    scheme = URIProperty("scheme", required=False)
    links = ArrayOfLinksProperty("links", required=False)

    def __hash__(self):
        """Calculate the hash.

        :return: Hash
        :rtype: int
        """
        return hash(
            (self.name, self.sort_as, self.code, self.scheme, _safe_tuple(self.links))
        )

    def __repr__(self):
        """Return a string representation of the object.

        :return: String representation
        :rtype: str
        """
        return "<Subject(name={}, sort_as={}, code={}, scheme={}, links={})>".format(
            self.name, self.sort_as, self.code, self.scheme, self.links
        )


class ArrayOfSubjectsProperty(BaseArrayProperty):
    """Property containing information about subjects.

    For example:
        - "Juvenile Fiction"
        - {
            name: "Juvenile Fiction"
          }
        - [
            "Juvenile Fiction",
            "Biography"
          ]
        - [
            {
                name: "Juvenile Fiction"
            },
            {
                name: "Biography"
            }
          ]
    """

    PARSER = AnyOfParser(
        [
            StringParser(),
            ArrayParser(AnyOfParser([StringParser(), TypeParser(Subject)])),
            TypeParser(Subject),
        ]
    )

    def __init__(self, key, required):
        """Initialize a new instance of ArrayOfSubjectsProperty class.

        :param key: Property's key
        :type key: str

        :param required: Boolean value indicating whether the property is required or not
        :type required: bool
        """
        super().__init__(key, required, self.PARSER, list, [])


class Owner(Node, PropertiesGrouping):
    """Object containing information about the collection's owners."""

    collection = ArrayOfContributorsProperty("collection", required=False)
    series = ArrayOfContributorsProperty("series", required=False)

    def __hash__(self):
        """Calculate the hash.

        :return: Hash
        :rtype: int
        """
        return hash((_safe_tuple(self.collection), _safe_tuple(self.series)))

    def __repr__(self):
        """Return a string representation of the object.

        :return: String representation
        :rtype: str
        """
        return "<Owner(collection={}, series={})>".format(
            encode(self.collection), encode(self.series)
        )


class Metadata(Node):
    """Dictionary containing manifest's metadata."""

    identifier = URIProperty("identifier", required=False)
    type = URIProperty("@type", required=False)
    title = LocalizableStringProperty("title", required=True)
    subtitle = LocalizableStringProperty("subtitle", required=False)
    modified = DateTimeProperty("modified", required=False)
    published = DateOrTimeProperty("published", required=False)
    languages = ListOfLanguagesProperty("language", required=False)
    sort_as = StringProperty("sortAs", required=False)
    authors = ArrayOfContributorsProperty("author", required=False)
    translators = ArrayOfContributorsProperty("translator", required=False)
    editors = ArrayOfContributorsProperty("editor", required=False)
    artists = ArrayOfContributorsProperty("artist", required=False)
    illustrators = ArrayOfContributorsProperty("illustrator", required=False)
    letterers = ArrayOfContributorsProperty("letterer", required=False)
    pencilers = ArrayOfContributorsProperty("penciler", required=False)
    colorists = ArrayOfContributorsProperty("colorist", required=False)
    inkers = ArrayOfContributorsProperty("inker", required=False)
    narrators = ArrayOfContributorsProperty("narrator", required=False)
    contributors = ArrayOfContributorsProperty("contributor", required=False)
    publishers = ArrayOfContributorsProperty("publisher", required=False)
    imprints = ArrayOfContributorsProperty("imprint", required=False)
    subjects = ArrayOfSubjectsProperty("subject", required=False)
    reading_progression = EnumProperty(
        "readingProgression",
        required=False,
        items=["rtl", "ltr", "ttb", "btt", "auto"],
        default_value="auto",
    )
    description = StringProperty("description", required=False)
    duration = NumberProperty("duration", required=False, exclusive_minimum=0)
    number_of_pages = NumberProperty(
        "numberOfPages", required=False, exclusive_minimum=0
    )
    belongs_to = TypeProperty("belongsTo", required=False, nested_type=Owner)
    isbns = ArrayOfStringsProperty("schema:isbn", required=False)
    issns = ArrayOfStringsProperty("schema:issn", required=False)
    reference_identifiers = ArrayOfURIsProperty("schema:identifier", required=False)

    def __init__(
        self,
        title=None,
        identifier=None,
        subtitle=None,
        modified=None,
        published=None,
        languages=None,
        sort_as=None,
        authors=None,
        translators=None,
        editors=None,
        artists=None,
        illustrators=None,
        letterers=None,
        pencilers=None,
        colorists=None,
        inkers=None,
        narrators=None,
        contributors=None,
        publishers=None,
        imprints=None,
        subjects=None,
        description=None,
        duration=None,
        number_of_pages=None,
        belongs_to=None,
        isbns=None,
        issns=None,
        reference_identifiers=None,
    ):
        """Initialize a new instance of Metadata class."""
        super().__init__()
        self.title = title
        self.identifier = identifier
        self.title = title
        self.subtitle = subtitle
        self.modified = modified
        self.published = published
        self.languages = languages
        self.sort_as = sort_as
        self.authors = authors
        self.translators = translators
        self.editors = editors
        self.artists = artists
        self.illustrators = illustrators
        self.letterers = letterers
        self.pencilers = pencilers
        self.colorists = colorists
        self.inkers = inkers
        self.narrators = narrators
        self.contributors = contributors
        self.publishers = publishers
        self.imprints = imprints
        self.subjects = subjects
        self.description = description
        self.duration = duration
        self.number_of_pages = number_of_pages
        self.belongs_to = belongs_to
        self.isbns = isbns
        self.issns = issns
        self.reference_identifiers = reference_identifiers

    def __eq__(self, other):
        """Compare two Metadata objects.

        :param other: Metadata object
        :type other: Metadata

        :return: Boolean value indicating whether two items are equal
        :rtype: bool
        """
        if not isinstance(other, Metadata):
            return False

        return (
            self.identifier == other.identifier
            and self.type == other.type
            and self.title == other.title
            and self.subtitle == other.subtitle
            and self.modified == other.modified
            and self.published == other.published
            and self.languages == other.languages
            and self.sort_as == other.sort_as
            and self.authors == other.authors
            and self.translators == other.translators
            and self.editors == other.editors
            and self.artists == other.artists
            and self.illustrators == other.illustrators
            and self.letterers == other.letterers
            and self.pencilers == other.pencilers
            and self.colorists == other.colorists
            and self.inkers == other.inkers
            and self.narrators == other.narrators
            and self.contributors == other.contributors
            and self.publishers == other.publishers
            and self.imprints == other.imprints
            and self.subjects == other.subjects
            and self.description == other.description
            and self.duration == other.duration
            and self.number_of_pages == other.number_of_pages
            and self.belongs_to == other.belongs_to
            and self.isbns == other.isbns
            and self.issns == other.issns
            and self.reference_identifiers == other.reference_identifiers
        )

    def __hash__(self):
        """Calculate the hash.

        :return: Hash
        :rtype: int
        """
        return hash(
            (
                self.identifier,
                self.type,
                self.title,
                self.subtitle,
                self.modified,
                self.published,
                _safe_tuple(self.languages),
                self.sort_as,
                _safe_tuple(self.authors),
                _safe_tuple(self.translators),
                _safe_tuple(self.editors),
                _safe_tuple(self.artists),
                _safe_tuple(self.illustrators),
                _safe_tuple(self.letterers),
                _safe_tuple(self.pencilers),
                _safe_tuple(self.colorists),
                _safe_tuple(self.inkers),
                _safe_tuple(self.narrators),
                _safe_tuple(self.contributors),
                _safe_tuple(self.publishers),
                _safe_tuple(self.imprints),
                _safe_tuple(self.subjects),
                self.description,
                self.duration,
                self.number_of_pages,
                self.belongs_to,
                tuple(self.isbns),
                tuple(self.issns),
                tuple(self.reference_identifiers),
            )
        )


class PresentationMetadata(Metadata):
    """RWPM extension containing presentation metadata."""

    clipped = BooleanProperty("clipped", False)
    continuous = BooleanProperty("continuous", False)
    fit = EnumProperty("fit", False, ["width", "height", "contain", "cover"])
    orientation = EnumProperty("orientation", False, ["auto", "landscape", "portrait"])
    overflow = EnumProperty(
        "overflow", False, ["auto", "paginated", "scrolled", "scrolled-continuous"]
    )
    spread = EnumProperty("spread", False, ["auto", "both", "none", "landscape"])

    def __eq__(self, other):
        """Compare two PresentationMetadata objects.

        :param other: PresentationMetadata object
        :type other: PresentationMetadata

        :return: Boolean value indicating whether two items are equal
        :rtype: bool
        """
        if not super().__eq__(other):
            return False

        if not isinstance(other, PresentationMetadata):
            return False

        return (
            self.clipped == other.clipped
            and self.continuous == other.continuous
            and self.fit == other.fit
            and self.orientation == other.orientation
            and self.overflow == other.overflow
            and self.spread == other.spread
        )

    def __hash__(self):
        """Calculate the hash.

        :return: Hash
        :rtype: int
        """
        return hash(
            (
                super().__hash__(),
                self.clipped,
                self.continuous,
                self.fit,
                self.orientation,
                self.overflow,
                self.spread,
            )
        )


class CompactCollection(Node):
    """A compact collection is defined as a grouping of links."""

    links = ArrayOfLinksProperty(key="links", required=True)

    def __init__(self, role=None, links=None):
        """Initialize a new instance of Collection class.

        :param role: Collection's roles (can be empty when self is a manifest)
        :type role: Optional[CollectionRole]

        :param links: Collection's links
        :type links: Optional[LinksList]
        """
        super().__init__()

        self._role = role
        self.links = links

    def __eq__(self, other):
        """Compare two CompactCollection objects.

        :param other: CompactCollection object
        :type other: CompactCollection

        :return: Boolean value indicating whether two items are equal
        :rtype: bool
        """
        if not isinstance(other, CompactCollection):
            return False

        return self.role == other.role and self.links == other.links

    def __hash__(self):
        """Calculate the hash.

        :return: Hash
        :rtype: int
        """
        return hash((self.role, self.links))

    @property
    def role(self):
        """Return the collection's role.

        :return: Collection's role.
        :rtype: Optional[CollectionRole]
        """
        return self._role


class Collection(CompactCollection):
    """A collection is defined as a grouping of metadata, links and sub-collections."""

    metadata = TypeProperty(
        key="metadata", required=True, nested_type=PresentationMetadata
    )

    def __init__(self, role=None, links=None, metadata=None):
        """Initialize a new instance of Collection class.

        :param role: Collection's role (can be empty when self is a manifest)
        :type role: Optional[CollectionRole]

        :param links: Collection's links
        :type links: Optional[LinksList]

        :param metadata: Collection's metadata
        :type metadata: Optional[Metadata]
        """
        super().__init__(role, links)

        self._role = role
        self._sub_collections = CollectionList()
        self.metadata = metadata

    def __eq__(self, other):
        """Compare two Collection objects.

        :param other: Collection object
        :type other: Collection

        :return: Boolean value indicating whether two items are equal
        :rtype: bool
        """
        if not super().__eq__(other):
            return False

        if not isinstance(other, Collection):
            return False

        return (
            self.metadata == other.metadata
            and self.sub_collections == other.sub_collections
        )

    def __hash__(self):
        """Calculate the hash.

        :return: Hash
        :rtype: int
        """
        return hash((super().__hash__(), self.metadata))

    @property
    def sub_collections(self):
        """Return a list of sub-collections.

        :return: List of sub-collections.
        :rtype: CollectionList
        """
        return self._sub_collections

    @property
    def compact(self):
        """Return a boolean value indicating if this collection is compact.

        :return: Boolean value indicating if this collection is compact
        :rtype: bool
        """
        return self.metadata is None and len(self._sub_collections) == 0

    @property
    def full(self):
        """Return a boolean value indicating if this collection is full.

        :return: Boolean value indicating if this collection is full
        :rtype: bool
        """
        return self.metadata is not None and len(self._sub_collections) > 0


class CollectionList(Node, list):
    """List of sub-collections."""

    def __init__(self, items=None):
        """Initialize a new instance of CollectionList class.

        :param items: (Optional) Items to be added to the list
        :type items: Optional[List]
        """
        super().__init__()

        if items is not None:
            if not isinstance(items, list):
                raise ValueError(f"Argument 'items' must be an instance of {list}")

            self.extend(items)

    def __hash__(self):
        """Calculate the hash.

        :return: Hash
        :rtype: int
        """
        return hash(tuple(self))

    def get_by_role(self, role):
        """Return collections with the specific role.

        :param role: Collection's role
        :type role: str

        :return: Collections with the specific role
        :rtype: List[Collection]
        """
        return [collection for collection in self if collection.role.key == role]


class CompactCollectionProperty(Property):
    """Property allowing to contain a compact sub-collection."""

    def __init__(self, key, required, role, collection_class=CompactCollection):
        """Initialize a new instance of CompactCollectionProperty class.

        :param key: Property's key
        :type key: str

        :param required: Boolean value indicating whether the property is required or not
        :type required: bool

        :param role: Collection role
        :type role: CollectionRole
        """
        if not isinstance(role, CollectionRole):
            raise ValueError(f"Argument 'role' must be an instance of {CollectionRole}")
        if not issubclass(collection_class, CompactCollection):
            raise ValueError(
                "Argument 'collection_class' must be a subclass of {}".format(
                    CompactCollection
                )
            )

        super().__init__(key, required, TypeParser(collection_class))

        self._role = role

    @property
    def role(self):
        """Return the sub-collection's role.

        :return: Sub-collection's role
        :rtype: CollectionRole
        """
        return self._role


class ArrayOfCollectionsProperty(BaseArrayProperty):
    """Property allowing to contain a compact sub-collection."""

    def __init__(self, key, required, role, collection_type=Collection):
        """Initialize a new instance of ArrayOfCollectionsProperty class.

        :param key: Property's key
        :type key: str

        :param required: Boolean value indicating whether the property is required or not
        :type required: bool

        :param role: Collection role
        :type role: CollectionRole
        """
        if not isinstance(role, CollectionRole):
            raise ValueError(f"Argument 'role' must be an instance of {CollectionRole}")
        if not issubclass(collection_type, Collection):
            raise ValueError(
                f"Argument 'collection_type' must be a subclass of {Collection}"
            )

        super().__init__(
            key,
            required,
            ArrayParser(TypeParser(collection_type), True),
            CollectionList,
        )

        self._role = role

    @property
    def role(self):
        """Return the sub-collection's role.

        :return: Sub-collection's role
        :rtype: CollectionRole
        """
        return self._role


class Manifestlike(Collection):
    """Base class for Manifest (Readium Web Publication Manifest) and Feed (OPDS 2).

    An OPDS 2 feed is defined as a RWPM with enumerated exceptions.

    This class implements the behavior common to both specs.  The
    alternative is to have the Feed class subclass Manifest and then
    implement a lot of exceptions.
    """



from webpub_manifest_parser.core import ManifestParser, ManifestParserFactory
from webpub_manifest_parser.odl.semantic import ODLSemanticAnalyzer
from webpub_manifest_parser.odl.syntax import ODLSyntaxAnalyzer
from webpub_manifest_parser.opds2.registry import (
    OPDS2CollectionRolesRegistry,
    OPDS2LinkRelationsRegistry,
    OPDS2MediaTypesRegistry,
)


class ODLFeedParserFactory(ManifestParserFactory):
    """Factory creating a new ODL parser."""

    def create(self):
        """Create a new ODL parser.

        :return: ODL parser
        :rtype: Parser
        """
        media_types_registry = OPDS2MediaTypesRegistry()
        link_relations_registry = OPDS2LinkRelationsRegistry()
        collection_roles_registry = OPDS2CollectionRolesRegistry()
        syntax_analyzer = ODLSyntaxAnalyzer()
        semantic_analyzer = ODLSemanticAnalyzer(
            media_types_registry, link_relations_registry, collection_roles_registry
        )
        parser = ManifestParser(syntax_analyzer, semantic_analyzer)

        return parser
